%% From mikexcohen.com
% phase-based connectivity
% code ammended by John Allen and also includes code from Chapter26 

load '../sampleEEGdata.mat'

%% setup some stuff and junk

% names of the channels you want to compute connectivity between
channel1 = 'p1';
channel2 = 'pz';


% create complex Morlet wavelet
center_freq = 5;
time      = -1:1/EEG.srate:1;
wavelet   = exp(2*1i*pi*center_freq.*time) .* exp(-time.^2./(2*(4/(2*pi*center_freq))^2));
half_wavN = (length(time)-1)/2;

% FFT parameters
n_wavelet = length(time);
n_data    = EEG.pnts;
n_conv    = n_wavelet+n_data-1;

% FFT of wavelet
waveletX = fft(wavelet,n_conv);
waveletX = waveletX ./ max(waveletX);

% initialize output time-frequency data
phase_data = zeros(2,EEG.pnts);
real_data  = zeros(2,EEG.pnts);

% find channel indices
chan1idx = find(strcmpi(channel1,{EEG.chanlocs.labels}));
chan2idx = find(strcmpi(channel2,{EEG.chanlocs.labels}));


% analytic signal of channel for epoch 1
fft_data = fft(squeeze(EEG.data(chan1idx,:,1)),n_conv);
as = ifft(waveletX.*fft_data,n_conv);
as = as(half_wavN+1:end-half_wavN);

% collect real and phase data
phase_data(1,:) = angle(as);
real_data(1,:)  = real(as);

% analytic signal of channel 1
fft_data = fft(squeeze(EEG.data(chan2idx,:,1)),n_conv);
as = ifft(waveletX.*fft_data,n_conv);
as = as(half_wavN+1:end-half_wavN);

% extract real and phase data  
% IMPORTANT: if you don't know what is going on here, we better review!
phase_data(2,:) = angle(as);
real_data(2,:)  = real(as);

%% setup figure and define plot handles

% open and name figure
figure, set(gcf,'Name','Movie magic minimizes the magic.');

% draw the filtered signals
subplot(221)
filterplotH1 = plot(EEG.times(1),real_data(1,1),'b');
hold on
filterplotH2 = plot(EEG.times(1),real_data(2,1),'m');
set(gca,'xlim',[EEG.times(1) EEG.times(end)],'ylim',[min(real_data(:)) max(real_data(:))])
xlabel('Time (ms)')
ylabel('Voltage (\muV)')
title([ 'Filtered signal at ' num2str(center_freq) ' Hz' ])

% draw the phase angle time series
subplot(222)
phaseanglesH1 = plot(EEG.times(1),phase_data(1,1),'b');
hold on
phaseanglesH2 = plot(EEG.times(1),phase_data(2,1),'m');
set(gca,'xlim',[EEG.times(1) EEG.times(end)],'ylim',[-pi pi]*1.1)
xlabel('Time (ms)')
ylabel('Phase angle (radian)')
title('Phase angle time series')

% draw phase angles in polar space
subplot(223)
polar2chanH1 = polar([zeros(1,1) phase_data(1,1)]',repmat([0 1],1,1)','b');
hold on
polar2chanH2 = polar([zeros(1,1) phase_data(2,1)]',repmat([0 1],1,1)','m');
title('Phase angles from two channels')

% draw phase angle differences in polar space
subplot(224)
polarAngleDiffH = polar([zeros(1,1) phase_data(2,1)-phase_data(1,1)]',repmat([0 1],1,1)','k');
title('Phase angle differences from two channels')

%% now update plots at each timestep

for ti=1:5:EEG.pnts
    
    % update filtered signals
    set(filterplotH1,'XData',EEG.times(1:ti),'YData',real_data(1,1:ti))
    set(filterplotH2,'XData',EEG.times(1:ti),'YData',real_data(2,1:ti))
    
    % update cartesian plot of phase angles
    set(phaseanglesH1,'XData',EEG.times(1:ti),'YData',phase_data(1,1:ti))
    set(phaseanglesH2,'XData',EEG.times(1:ti),'YData',phase_data(2,1:ti))
    
    subplot(223)
    cla
    polar([zeros(1,ti) phase_data(1,1:ti)]',repmat([0 1],1,ti)','b');
    hold on
    polar([zeros(1,ti) phase_data(2,1:ti)]',repmat([0 1],1,ti)','m');
    
    subplot(224)
    cla
    polar([zeros(1,ti) phase_data(2,1:ti)-phase_data(1,1:ti)]',repmat([0 1],1,ti)','k');
    
    drawnow
end


%% Assignment
% Choose different channels that are farther apart and rerun the
% visualization


%% now quantify phase synchronization between the two channels

% phase angle differences
phase_angle_differences = phase_data(2,:)-phase_data(1,:);

% euler representation of angles
euler_phase_differences = exp(1i*phase_angle_differences);

% mean vector (in complex space)
mean_complex_vector = mean(euler_phase_differences);

% length of mean vector (this is the "M" from Me^ik, and is the measure of phase synchronization)
phase_synchronization = abs(mean_complex_vector);

disp([ 'Synchronization between ' channel1 ' and ' channel2 ' is ' num2str(phase_synchronization) '!' ])

% of course, this could all be done on one line:
phase_synchronization = abs(mean(exp(1i*(phase_data(2,:)-phase_data(1,:)))));

% notice that the order of subtraction is meaningless (see below), which means that this measure of synchronization is non-directional!
phase_synchronization_backwards = abs(mean(exp(1i*(phase_data(1,:)-phase_data(2,:)))));


% now plot mean vector
subplot(224)
hold on
h=polar([0 angle(mean_complex_vector)],[0 phase_synchronization]);
set(h,'linewidth',6,'color','g')

%% phase clustering is phase-invariant

figure(2), clf
subplot(221)
polar(repmat(phase_data(2,:)-phase_data(1,:),1,2)',repmat([0 1],1,EEG.pnts)','k');
title([ 'Phase synchronization: ' num2str(abs(mean(exp(1i*(diff(phase_data,1)))))) ])

new_phase_data = phase_data;
for i=2:4
    subplot(2,2,i)
    
    % add random phase offset
    new_phase_data(1,:) = new_phase_data(1,:)+rand*pi;
    
    % plot again
    polar(repmat(new_phase_data(2,:)-new_phase_data(1,:)+pi/2,1,2)',repmat([0 1],1,EEG.pnts)','k');
    title([ 'Phase synchronization: ' num2str(abs(mean(exp(1i*(diff(new_phase_data,1)))))) ])
end

%% 

%% ISPC over time vs. over trials

% clear workspace and reload data
clear all;
close all;
clc;

load '../sampleEEGdata.mat'

% names of the channels you want to compute connectivity between
channel1 = 'p1';
channel2 = 'pz';

% find channel indices
chan1idx = find(strcmpi(channel1,{EEG.chanlocs.labels}));
chan2idx = find(strcmpi(channel2,{EEG.chanlocs.labels}));


% Wavelet and FFT parameters
center_freq = 5;
time      = -1:1/EEG.srate:1;
wavelet   = exp(2*1i*pi*center_freq.*time) .* exp(-time.^2./(2*(4/(2*pi*center_freq))^2));
half_wavN = (length(time)-1)/2;

n_wavelet = length(time);
n_data    = EEG.pnts*EEG.trials;
n_conv    = n_wavelet+n_data-1;

% initialize output time-frequency data
phase_data = zeros(2,EEG.pnts,EEG.trials);

% FFT of wavelet 
waveletX = fft(wavelet,n_conv);
waveletX = waveletX ./ max(waveletX);


% analytic signal of channel 1
fft_data = fft(reshape(EEG.data(chan1idx,:,:),1,[]),n_conv);
as = ifft(waveletX.*fft_data,n_conv);
as = as(half_wavN+1:end-half_wavN);
as = reshape(as,EEG.pnts,EEG.trials);

% collect real and phase data
phase_data(1,:,:) = angle(as);

% analytic signal of channel 2
fft_data = fft(reshape(EEG.data(chan2idx,:,:),1,[]),n_conv);
as = ifft(waveletX.*fft_data,n_conv);
as = as(half_wavN+1:end-half_wavN);
as = reshape(as,EEG.pnts,EEG.trials);

% collect real and phase data
phase_data(2,:,:) = angle(as);


figure(3), clf

% ISPC over trials
subplot(211)
ispc_trials = abs(mean(exp(1i*diff(phase_data,[],1)),3));
plot(EEG.times,ispc_trials)
set(gca,'xlim',[-200 1200])
xlabel('Time (ms)'), ylabel('ISPC')

% ISPC over time
subplot(212)
ispc_time = squeeze(abs(mean(exp(1i*diff(phase_data,[],1)),2)));
plot(1:EEG.trials,ispc_time)
xlabel('Trials'), ylabel('ISPC')


%% Figure 26.9 -- PLI vs ISPC using simulatd phase angles

    % number of "trials"
    n = 100;

    % Get info on sign
    doc sign

    %% now make the plots
    figure

    subplot(221)
    phases = rand(n,1)*pi;
    polar([phases; phases],repmat([0 1]',n,1),'k');
    pli  = abs(mean(sign(imag(exp(1i*phases)))));
    ispc = abs(mean(exp(1i*phases)));
    title([ 'PLI=' num2str(pli) ', ISPC=' num2str(ispc) ])

    subplot(222)
    phases = phases-pi/2;
    polar([phases; phases],repmat([0 1]',n,1),'k');
    pli  = abs(mean(sign(imag(exp(1i*phases)))));
    ispc = abs(mean(exp(1i*phases)));
    title([ 'PLI=' num2str(pli) ', ISPC=' num2str(ispc) ])


    subplot(223)
    phases = rand(n,1)/2+pi/3+.25;
    polar([phases; phases],repmat([0 1]',n,1),'k');
    pli  = abs(mean(sign(imag(exp(1i*phases)))));
    ispc = abs(mean(exp(1i*phases)));
    title([ 'PLI=' num2str(pli) ', ISPC=' num2str(ispc) ])

    subplot(224)
    phases = phases-pi/2;
    polar([phases; phases],repmat([0 1]',n,1),'k');
    pli  = abs(mean(sign(imag(exp(1i*phases)))));
    ispc = abs(mean(exp(1i*phases)));
    title([ 'PLI=' num2str(pli) ', ISPC=' num2str(ispc) ])


%% Figure 26.10 -- ISPC and dwPLI with real data, over time and over trials across frequencies

load '../sampleEEGdata.mat'

% select channels
channel1 = 'fp1';
channel2 = 'fp2';

% specify some time-frequency parameters
freqs2use  = logspace(log10(4),log10(30),15); % 4-30 Hz in 15 steps
times2save = -400:10:800;
timewindow = linspace(1.5,3,length(freqs2use)); % number of cycles on either end of the center point (1.5 means a total of 3 cycles))
baselinetm = [-400 -200];

% wavelet and FFT parameters
time          = -1:1/EEG.srate:1;
half_wavelet  = (length(time)-1)/2;
num_cycles    = logspace(log10(4),log10(8),length(freqs2use));
n_wavelet     = length(time);
n_data        = EEG.pnts*EEG.trials;
n_convolution = n_wavelet+n_data-1;

% time in indices
times2saveidx = dsearchn(EEG.times',times2save');
baselineidx   = dsearchn(times2save',baselinetm');

chanidx = zeros(1,2); % always initialize!
chanidx(1) = find(strcmpi(channel1,{EEG.chanlocs.labels}));
chanidx(2) = find(strcmpi(channel2,{EEG.chanlocs.labels}));

% data FFTs
data_fft1 = fft(reshape(EEG.data(chanidx(1),:,:),1,n_data),n_convolution);
data_fft2 = fft(reshape(EEG.data(chanidx(2),:,:),1,n_data),n_convolution);

% initialize
ispc    = zeros(length(freqs2use),EEG.pnts);  % Inter-site Phase clustering
pli     = zeros(length(freqs2use),EEG.pnts);  % Phase Lag Index 
wpli    = zeros(length(freqs2use),EEG.pnts);  % Weighted Phase Lag Index (weighted by distance from the real axis)
dwpli   = zeros(length(freqs2use),EEG.pnts);  % debiased weighted phase-lag index 
dwpli_t = zeros(length(freqs2use),length(times2save)); % debiased weighted phase-lag index over time
ispc_t  = zeros(length(freqs2use),length(times2save));  % Inter-site Phase clustering over time

for fi=1:length(freqs2use)  % loop on frequencies
    
    % create wavelet and take FFT
    s = num_cycles(fi)/(2*pi*freqs2use(fi));
    wavelet_fft = fft( exp(2*1i*pi*freqs2use(fi).*time) .* exp(-time.^2./(2*(s^2))) ,n_convolution);
    
    % phase angles from channel 1 via convolution
    convolution_result_fft = ifft(wavelet_fft.*data_fft1,n_convolution);
    convolution_result_fft = convolution_result_fft(half_wavelet+1:end-half_wavelet);
    sig1 = reshape(convolution_result_fft,EEG.pnts,EEG.trials);
    
    % phase angles from channel 2 via convolution
    convolution_result_fft = ifft(wavelet_fft.*data_fft2,n_convolution);
    convolution_result_fft = convolution_result_fft(half_wavelet+1:end-half_wavelet);
    sig2 = reshape(convolution_result_fft,EEG.pnts,EEG.trials);
    
    % cross-spectral density
    cdd = sig1 .* conj(sig2);  % sig1.*conj(sig1) is the same thing as abs(sig1).^2, but faster; and do this with two signals!
                               % With two signals, equivalent to and faster than the Euler notation: 
                               % abs(sig1).*abs(sig2).*exp(1i*(angle(sig1)-angle(sig2))).
                               
    % ISPC
    ispc(fi,:) = abs(mean(exp(1i*angle(cdd)),2)); % note: equivalent to ispc(fi,:) = abs(mean(exp(1i*(angle(sig1)-angle(sig2))),2));
    
    
    % take imaginary part of signal only
    cdi = imag(cdd);
    
    % phase-lag index
    pli(fi,:)  = abs(mean(sign(imag(cdd)),2));  % one value for every timepoint
    
    % weighted phase-lag index (eq. 8 in Vink et al. NeuroImage 2011)
    wpli(fi,:) = abs( mean( abs(cdi).*sign(cdi) ,2) )./mean(abs(cdi),2);
    
    % debiased weighted phase-lag index (shortcut, as implemented in fieldtrip)
    imagsum      = sum(cdi,2);
    imagsumW     = sum(abs(cdi),2);
    debiasfactor = sum(cdi.^2,2);
    dwpli(fi,:)  = (imagsum.^2 - debiasfactor)./(imagsumW.^2 - debiasfactor);
    
    % compute time window in indices for this frequency
    time_window_idx = round((1000/freqs2use(fi))*timewindow(fi)/(1000/EEG.srate));

    for ti=1:length(times2save)
        imagsum        = sum(cdi(times2saveidx(ti)-time_window_idx:times2saveidx(ti)+time_window_idx,:),1);
        imagsumW       = sum(abs(cdi(times2saveidx(ti)-time_window_idx:times2saveidx(ti)+time_window_idx,:)),1);
        debiasfactor   = sum(cdi(times2saveidx(ti)-time_window_idx:times2saveidx(ti)+time_window_idx,:).^2,1);
        dwpli_t(fi,ti) = mean((imagsum.^2 - debiasfactor)./(imagsumW.^2 - debiasfactor));

        % compute phase synchronization
        phasesynch     = abs(mean(exp(1i*angle(cdd(times2saveidx(ti)-time_window_idx:times2saveidx(ti)+time_window_idx,:))),1));
        ispc_t(fi,ti)  = mean(phasesynch);
    end
end

% baseline subtraction from all measures
ispc    = bsxfun(@minus,ispc,mean(ispc(:,baselineidx(1):baselineidx(2)),2)); % not plotted in the book, but you can plot it for comparison with PLI
ispc_t  = bsxfun(@minus,ispc_t,mean(ispc_t(:,baselineidx(1):baselineidx(2)),2));
pli     = bsxfun(@minus,pli,mean(pli(:,baselineidx(1):baselineidx(2)),2));
dwpli   = bsxfun(@minus,dwpli,mean(dwpli(:,baselineidx(1):baselineidx(2)),2));
dwpli_t = bsxfun(@minus,dwpli_t,mean(dwpli_t(:,baselineidx(1):baselineidx(2)),2));

figure
subplot(311)
contourf(times2save,freqs2use,ispc(:,times2saveidx),40,'linecolor','none')
set(gca,'clim',[-.2 .2],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
title('ISPC over trials')

subplot(312)
contourf(times2save,freqs2use,pli(:,times2saveidx),40,'linecolor','none')
set(gca,'clim',[-.2 .2],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
title('PLI over trials')

subplot(313)
contourf(times2save,freqs2use,dwpli(:,times2saveidx),40,'linecolor','none')
set(gca,'clim',[-.2 .2],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
title('dWPLI over trials')

% subplot(232)
% contourf(times2save,freqs2use,ispc_t,40,'linecolor','none')
% set(gca,'clim',[-.1 .1],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
% title('ICPS over time')
% 
% subplot(234)
% contourf(times2save,freqs2use,pli_t,40,'linecolor','none')
% set(gca,'clim',[-.1 .1],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
% title('PLI over time')
% 
% subplot(236)
% contourf(times2save,freqs2use,dwpli_t,40,'linecolor','none')
% set(gca,'clim',[-.1 .1],'yscale','log','ytick',round(logspace(log10(freqs2use(1)),log10(freqs2use(end)),8)))
% title('dWPLI over time')


%% Figure 26.11 -- ISPC vs dwPLI over time -- A movie!

trial2plot = 10; % any trial between 1 and 99 (book uses trial 10)
center_freq = 4.6; % Hz (book uses 4.6)


% create wavelet and take FFT
s = 4.5/(2*pi*center_freq);
wavelet_fft = fft( exp(2*1i*pi*center_freq.*time) .* exp(-time.^2./(2*(s^2))) ,n_convolution);
% phase angles from channel 1 via convolution
convolution_result_fft = ifft(wavelet_fft.*data_fft1,n_convolution);
convolution_result_fft = convolution_result_fft(half_wavelet+1:end-half_wavelet);
sig1 = reshape(convolution_result_fft,EEG.pnts,EEG.trials);
% phase angles from channel 2 via convolution
convolution_result_fft = ifft(wavelet_fft.*data_fft2,n_convolution);
convolution_result_fft = convolution_result_fft(half_wavelet+1:end-half_wavelet);
sig2 = reshape(convolution_result_fft,EEG.pnts,EEG.trials);
% cross-spectral density
xsd  = sig1 .* conj(sig2);
xsdi = imag(xsd);

dwpli = zeros(size(EEG.times));
ispc  = zeros(size(EEG.times));

[junk,animate_start] = min(abs(EEG.times-0));
[junk,animate_stop]  = min(abs(EEG.times-1000));

time_window_idx = round(100*timewindow(1)/(1000/EEG.srate));

figure
subplot(121)
hpol = polar(repmat(angle(xsd(animate_start:animate_start+time_window_idx-1,trial2plot)),1,2)',[zeros(time_window_idx,1) ones(time_window_idx,1)]','k-o');
subplot(122)
hplo2 = plot(EEG.times,0,'r');
hold on
hplo1 = plot(EEG.times,0,'b');

for idx=animate_start:animate_stop
    
    % update angles
    for i=1:length(hpol)
        set(hpol(i),'XData',[0; cos(angle(xsd(idx+i,trial2plot)))],'YData',[0; sin(angle(xsd(idx+i,trial2plot)))]);
    end
    title([ num2str(round(EEG.times(idx))) '-' num2str(round(EEG.times(idx+i))) ' ms' ])

    % compute ICPS and dwPLI
    ispc(idx) = abs(mean(exp(1i*angle(xsd(idx:idx+i,trial2plot))),1));
    
    imagsum        = sum(xsdi(idx:idx+i,trial2plot),1);
    imagsumW       = sum(abs(xsdi(idx:idx+i,trial2plot)),1);
    debiasfactor   = sum(xsdi(idx:idx+i,trial2plot).^2,1);
    dwpli(idx) = mean((imagsum.^2 - debiasfactor)./(imagsumW.^2 - debiasfactor));
    
    
    set(hplo1,'XData',EEG.times(1:idx),'YData',ispc(1:idx));
    set(hplo2,'XData',EEG.times(1:idx),'YData',dwpli(1:idx));
    set(gca,'xlim',EEG.times([animate_start animate_stop]),'ylim',[-.1 1.1])
    axis square
    
    pause(0.01)
end
